@@ -10,7 +10,7 @@ import tempfile
 import configparser
 from subprocess import run, PIPE, DEVNULL
 
-version = "0.9.0"
+version = "0.9.1"
 
 
 class PassNode(urwid.AttrMap):
@@ -23,19 +23,25 @@ class PassNode(urwid.AttrMap):
         self.text = node if node else "-- EMPTY --"
         self.path = os.path.join(root, node) if node else ''
         self.icon = config.icon_dir if isdir else config.icon_file if node else ''
-        # topdown option in os.walk makes this possible,
-        # so that children folders are traversed before its parent
-        self.count = str(len(Pass.all_pass[self.path])) if isdir else ''
 
         super().__init__(urwid.Columns([
                 ('pack', urwid.Text(self.icon)),
                 urwid.Text(self.text, wrap='clip'),
-                ('pack', urwid.Text(self.count))
+                ('pack', urwid.Text(''))
             ]),
             'dir' if isdir else '' if node else 'bright',
             'focusdir' if isdir else 'focus' if node else 'bright',
         )
 
+        self.update_count()
+
+    def update_count(self):
+        # topdown option in os.walk makes this possible,
+        # so that children folders are traversed before its parent
+        if self.isdir:
+            count = len(Pass.all_pass[self.path])
+            self.original_widget.contents[2][0].set_text(str(count))
+
     def keypress(self, size, key):
         """ let the widget pass through the keys to parent widget """
         return key
@@ -144,17 +150,22 @@ class PassList(urwid.ListBox):
         self._ui.update_view()
 
     def insert(self, node):
-        passnode = PassNode(node, self.root)
-
-        # change stored list
-        root_list = Pass.all_pass[self.root]
-        if len(root_list) == 1 and root_list[0].node is None:
-            root_list.pop()
-        inserted_pos = root_list.insert_sorted(passnode)
+        def insert_relative(r, n):
+            n1, sep, n2 = n.partition(os.sep)
+            if sep == os.sep:
+                insert_relative(os.path.join(r, n1), n2)
+            passnode = PassNode(n1, r, isdir=(sep == os.sep))
+            # change stored list
+            if Pass.all_pass.get(r) is None:
+                Pass.all_pass[r] = FolderWalker(r)
+            pos = Pass.all_pass[r].insert_sorted(passnode)
+            # change saved cursor position
+            Pass.all_pass[r].pos = pos
+            return pos
 
+        inserted_pos = insert_relative(self.root, node)
         # change listwalker
-        self.body[:] = root_list
-
+        self.body[:] = Pass.all_pass[self.root]
         # focus the new node
         self.set_focus(inserted_pos)
 
@@ -162,19 +173,21 @@ class PassList(urwid.ListBox):
 
     def delete(self, pos):
         # change stored list
-        root_list = Pass.all_pass[self.root]
-        root_list.pop(pos)
-        if len(root_list) == 0:
-            root_list.append(PassNode(None, None))
-
+        Pass.all_pass[self.root].pop(pos)
         # change listwalker
-        self.body[:] = root_list
+        self.body[:] = Pass.all_pass[self.root]
 
         self._ui.update_view()
 
+    def update_root_count(self):
+        for n in Pass.all_pass[os.path.dirname(self.root)]:
+            if n.node == self.root and n.isdir:
+                n.update_count()
+                return
+
 
 class FolderWalker(list):
-    def __init__(self, root, dirs, files):
+    def __init__(self, root, dirs=[], files=[]):
         self.pos = 0  # cursor position
 
         self[:] = [PassNode(f, root, True) for f in sorted(dirs)] + \
@@ -184,11 +197,20 @@ class FolderWalker(list):
         if len(self) == 0:
             self[:] = [PassNode(None, None)]
 
+    def pop(self, index=-1):
+        super().pop(index)
+        if len(self) == 0:
+            super().append(PassNode(None, None))
+
     def insert_sorted(self, node):
         # if node already exist, return the index
-        node_list = [n.node for n in self]
-        if node.node in node_list:
-            return node_list.index(node.node)
+        for n in self:
+            if n.node == node.node and n.isdir == node.isdir:
+                return self.index(n)
+
+        # pop the empty placeholder node beforehand
+        if len(self) == 1 and self[0].node is None:
+            super().pop()
 
         # insert and sort, with directories sorted before files
         super().insert(self.pos, node)
@@ -197,9 +219,8 @@ class FolderWalker(list):
         return self.index(node)
 
 
-# TODO: update count
 # TODO: background preview, or/and cache preview results
-# TODO: focus changing when editing
+# TODO: deal with focus changing when editing
 class UI(urwid.Frame):
     def __init__(self):
         self._last_preview = None
@@ -321,10 +342,10 @@ class UI(urwid.Frame):
             self.unfocus_edit()
             self.search_in_dir(self._search_pattern, 1)
         elif self._edit_type == "generate":
-            # TODO: can not accept "dir/node" format
             self.unfocus_edit()
             self.run_pass(Pass.generate, self.listbox.insert,
                           self.editbox.edit_text, self.listbox.root, "Generate: {}")
+            self.listbox.update_root_count()
         elif self._edit_type == "insert":
             self._insert_node = self.editbox.edit_text
             self.focus_edit("insert_password", 'Enter password: ', '*')
@@ -338,6 +359,7 @@ class UI(urwid.Frame):
                 self.run_pass(Pass.insert, self.listbox.insert,
                               self._insert_node, self.listbox.root, "Insert: {}",
                               args=(self._insert_pass,))
+                self.listbox.update_root_count()
             else:
                 self.message("Password is not the same", alert=True)
 
@@ -363,9 +385,9 @@ class UI(urwid.Frame):
         node = self.listbox.focus.text
         path = os.path.join(self.listbox.root, node)
 
-        if not force and path == self._last_preview:
+        if not force and self.listbox.focus == self._last_preview:
             return
-        self._last_preview = path
+        self._last_preview = self.listbox.focus
 
         if self.listbox.focus.isdir:
             preview = "\n".join([(f.icon + f.text) for f in Pass.all_pass[path]])
@@ -394,6 +416,7 @@ class UI(urwid.Frame):
             self.run_pass(Pass.delete, self.listbox.delete,
                           self.listbox.focus.node, self.listbox.root,
                           "Deleting {}", largs=(self.listbox.focus_position,))
+            self.listbox.update_root_count()
         elif key in ['n', 'N']:
             self.message("Abort.")
         else:
@@ -471,6 +494,7 @@ class Pass:
     FALLBACK_PASS_DIR = os.path.join(os.getenv("HOME"), ".password_store")
     PASS_DIR = os.getenv("PASSWORD_STORE_DIR", FALLBACK_PASS_DIR)
     X_SELECTION = os.getenv("PASSWORD_STORE_X_SELECTION", "clipboard")
+    EDITOR = os.getenv("EDITOR", "vi")
     all_pass = dict()
 
     @classmethod
@@ -485,35 +509,48 @@ class Pass:
                 cls.all_pass[root] = FolderWalker(root, dirs, files)
 
     @staticmethod
-    def show(node):
-        result = run(['pass', 'show', node], stdout=PIPE, stderr=PIPE, text=True)
+    def show(path):
+        logging.debug("Showing password for {}".format(path))
+        result = run(['pass', 'show', path], stdout=PIPE, stderr=PIPE, text=True)
         return result
 
-    @staticmethod
-    def edit(node):
-        # can not pipe stdout because this will start vim
-        # TODO: work around by manually edit temp file and insert with multiline
-        result = run(['pass', 'edit', node], stderr=PIPE, text=True)
-        return result
+    @classmethod
+    def edit(cls, path):
+        # work around terminal output by manually edit temp file and insert with multiline
+        with tempfile.NamedTemporaryFile() as fp:
+            res = cls.show(path)
+            if res.returncode != 0:
+                return res
+            fp.write(res.stdout.encode())
+            fp.flush()
+            # can not pipe stdout because editor won't show otherwise
+            res = run([cls.EDITOR, fp.name], stderr=PIPE)
+            if res.returncode != 0:
+                return res
+            fp.seek(0)
+            password = fp.read()
+            res = run(['pass', 'insert', '-m', '-f', path], input=password,
+                      stderr=PIPE, stdout=PIPE)
+            return res
 
     @staticmethod
-    def insert(node, password):
+    def insert(path, password):
         pw = password + '\n' + password + '\n'
-        result = run(['pass', 'insert', '-f', node], input=pw,
+        result = run(['pass', 'insert', '-f', path], input=pw,
                      stdout=PIPE, stderr=PIPE, text=True)
         return result
 
     @staticmethod
-    def generate(node):
-        command = ['pass', 'generate', '-f', node]
+    def generate(path):
+        command = ['pass', 'generate', '-f', path]
         if config.no_symbols:
             command.append('-n')
         result = run(command, stdout=PIPE, stderr=PIPE, text=True)
         return result
 
     @staticmethod
-    def delete(node):
-        command = ['pass', 'rm', '-r', '-f', node]
+    def delete(path):
+        command = ['pass', 'rm', '-r', '-f', path]
         result = run(command, stdout=PIPE, stderr=PIPE, text=True)
         return result
 
